#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Multimap module.

puppeteer equivalent: lib/Multimap.js
"""


from typing import Any, Dict, List, Optional


class Multimap:
    """Multimap class."""

    def __init__(self) -> None:
        """Make new multimap."""
        # maybe defaultdict(set) is better
        self._map: Dict[Optional[str], List[Any]] = {}

    def set(self, key: Optional[str], value: Any) -> None:
        """Set value."""
        _set = self._map.get(key)
        if not _set:
            _set = []
            self._map[key] = _set
        if value not in _set:
            _set.append(value)

    def get(self, key: Optional[str]) -> List[Any]:
        """Get values."""
        return self._map.get(key, [])

    def has(self, key: Optional[str]) -> bool:
        """Check key is in this map."""
        return key in self._map

    def hasValue(self, key: Optional[str], value: Any) -> bool:
        """Check value is in this map."""
        _set = self._map.get(key, [])
        return value in _set

    def size(self) -> int:
        """Length of this map."""
        return len(self._map)

    def delete(self, key: Optional[str], value: Any) -> bool:
        """Delete value from key."""
        values = self.get(key)
        result = value in values
        if result:
            values.remove(value)
        if len(values) == 0:
            self._map.pop(key)
        return result

    def deleteAll(self, key: Optional[str]) -> None:
        """Delete all value of the key."""
        self._map.pop(key, None)

    def firstValue(self, key: Optional[str]) -> Any:
        """Get first value of the key."""
        _set = self._map.get(key)
        if not _set:
            return None
        return _set[0]

    def firstKey(self) -> Optional[str]:
        """Get first key."""
        return next(iter(self._map.keys()))

    def valuesArray(self) -> List[Any]:
        """Get all values as list."""
        result: List[Any] = []
        for values in self._map.values():
            result.extend(values)
        return result

    def clear(self) -> None:
        """Clear all entries of this map."""
        self._map.clear()



from zipfile import ZipFile


def validate_zip_file(zip_file_path, app_path):
    try:
        zip_file = ZipFile(zip_file_path)
    except Exception:
        raise Exception('Failed to read source files zip for validation.') from None

    files = [filename.strip('/') for filename in zip_file.namelist()]
    top_level_files = [filename for filename in files if len(filename.split('/')) == 1]

    if len(top_level_files) > 1:
        raise Exception('Please place all your source files in a single folder and zip it')

    # Check if .biolib folder is present
    if f'{top_level_files[0]}/.biolib' in files and f'{top_level_files[0]}/biolib' in files:
        raise Exception('You provided both a biolib and a .biolib folder. Please only provide the .biolib folder')

    elif f'{top_level_files[0]}/.biolib' in files:
        biolib_folder_path = f'{top_level_files[0]}/.biolib'

    elif f'{top_level_files[0]}/biolib' in files:
        raise Exception('Your biolib folder appears to be called "biolib" - it must be called ".biolib".')

    else:
        raise Exception(f'Could not find a .biolib folder in provided application folder {app_path}')

    # Check if the config file is present
    if f'{biolib_folder_path}/config.yml' in files:
        # TODO: Changed to make sense in CLI. Rewrite this check at some point
        pass

    elif f'{biolib_folder_path}/config.yaml' in files:
        raise Exception('Your biolib config file has the .yaml file extension - it must be .yml')

    else:
        raise Exception(
            f'Could not find config.yml file. Please provide a yaml file named config.yml in the .biolib folder'
        )



from enum import Enum

from biolib.typing_utils import TypedDict, List, Optional, Dict, Literal
from biolib.biolib_api_client.common_types import SemanticVersion


class AppVersionSlim(SemanticVersion):
    created_at: str
    public_id: str


class AppVersion(AppVersionSlim):
    app: str
    description: str
    is_runnable_by_user: bool
    source_code_license: str
    stdout_render_type: Literal['text', 'markdown']
    main_output_file: Optional[str]


class App(TypedDict):
    account_display_name: str
    account_handle: str
    account_id: str
    account_profile_picture: str
    active_version: str
    allow_client_side_execution: bool
    created_at: str
    is_license_required: bool
    name: str
    public_id: str
    state: str


class AppGetResponse(TypedDict):
    app: App
    app_uri: str
    app_version: AppVersion


class StdoutRenderType(Enum):
    MARKDOWN = 'markdown'
    TEXT = 'text'


class RemoteHost(TypedDict):
    hostname: str


class ModuleEnvironment(Enum):
    BIOLIB_APP = 'biolib-app'
    BIOLIB_CUSTOM = 'biolib-custom'
    BIOLIB_ECR = 'biolib-ecr'


class FilesMapping(TypedDict):
    from_path: str
    to_path: str


class LargeFileSystemMapping(TypedDict):
    presigned_download_url: str
    size_bytes: int
    to_path: str
    uuid: str


class _Module(TypedDict):
    command: str
    environment: Literal['biolib-app', 'biolib-custom', 'biolib-ecr']
    image_uri: str
    estimated_image_size_bytes: Optional[int]
    input_files_mappings: List[FilesMapping]
    large_file_systems: List[LargeFileSystemMapping]
    name: str
    output_files_mappings: List[FilesMapping]
    source_files_mappings: List[FilesMapping]
    working_directory: str


# type optional keys with total=False
class Module(_Module, total=False):
    secrets: Dict[str, str]


class _AppVersionOnJob(TypedDict):
    client_side_executable_zip: Optional[str]
    consumes_stdin: bool
    is_runnable_by_user: bool
    public_id: str
    remote_hosts: List[RemoteHost]
    settings: List[Dict]
    stdout_render_type: Literal['text', 'markdown']
    main_output_file: Optional[str]


# type optional keys with total=False
class AppVersionOnJob(_AppVersionOnJob, total=False):
    modules: List[Module]



# Imports to hide
from biolib import typing_utils as _typing_utils
from biolib.app import BioLibApp as _BioLibApp
from biolib.biolib_logging import logger as _logger, logger_no_user_data as _logger_no_user_data
from biolib.biolib_api_client import BiolibApiClient as _BioLibApiClient
from biolib.jobs import Job as _Job
from biolib import user as _user

import biolib.app
import biolib.cli
import biolib.utils


# ------------------------------------ Function definitions for public Python API ------------------------------------

def call_cli() -> None:
    biolib.cli.main()


def load(uri: str) -> _BioLibApp:
    return _BioLibApp(uri)


def get_job(job_id: str) -> _Job:
    return _Job(uuid=job_id)


def sign_in() -> None:
    _user.sign_in()


def sign_out() -> None:
    _user.sign_out()


def login() -> None:
    sign_in()


def logout() -> None:
    sign_out()


def set_api_base_url(api_base_url: str) -> None:
    _BioLibApiClient.initialize(base_url=api_base_url)


def set_api_token(api_token: str) -> None:
    api_client = _BioLibApiClient.get()
    api_client.sign_in_with_api_token(api_token)


def set_log_level(level: _typing_utils.Union[str, int]) -> None:
    _logger.setLevel(level)
    _logger_no_user_data.setLevel(level)


# -------------------------------------------------- Configuration ---------------------------------------------------
__version__ = biolib.utils.BIOLIB_PACKAGE_VERSION
_DEFAULT_LOG_LEVEL = 'INFO' if biolib.utils.IS_RUNNING_IN_NOTEBOOK else 'WARNING'
_logger.configure(default_log_level=_DEFAULT_LOG_LEVEL)
_logger_no_user_data.configure(default_log_level=_DEFAULT_LOG_LEVEL)

set_api_base_url(biolib.utils.BIOLIB_BASE_URL)
